import collections
import itertools
import typing

from .const import STRONGEST, STRONG, NEUTRAL, WEAK, WEAKEST
from .exception import NoMatch
from .group import ResourceGroup
from .item import ResourceItem
from .utils import unique_everseen, quote_str

__all__ = ['NoMatch', 'Resources', "ResourceGroup"]


RT = typing.TypeVar('RT', bound='Resources')


def split_resources(resources: str) -> typing.Generator[typing.Tuple[str, str], None, None]:
    """ Split a resources string into group-name, group-definition tuples

    :param resources: One or more resource definitions (either multi-line or semi-colon separated),
                      including group names
    :raises :py:exc:`ValueError`: if the string is an invalid resources definition
    :return: A generator yielding group-name, group definition tuples
    """
    key = None
    current = ''
    whitespace = ''
    quote = None
    escape = False  # True: include next character unconditionally

    resources = resources.strip()
    if resources:
        for c in resources+'\n':
            if escape:
                # This character is escaped, so include it unconditionally
                escape = False
                current += c
            elif c == '\\':
                # Escape next character. Keep escape character in values
                if key is not None:
                    current += c
                escape = True
            elif quote:
                # Inside a quoted string. Quotes in keys should be removed, but in values they need to be kept
                if c == quote:
                    # End of string
                    if key is not None:
                        current += c
                    quote = None
                else:
                    # Include any other character
                    current += c
            elif c in (' ', '\t'):
                # Whitespace. Ignore leading whitespace, collect all other whitespace in a separate variable
                if current:
                    whitespace += c
            elif c in ('"', "'"):
                # Start of a quoted string. Also include whitespace into current
                quote = c
                current += whitespace
                if key is not None:
                    current += c
                whitespace = ''
            elif c in (';', '\n'):
                # Field separator. Ignore trailing whitespace, reset state
                if key is None:
                    raise ValueError(f"missing key in {current!r}")

                yield key, whitespace + current
                key = None
                current = ''
                whitespace = ''
            elif key is None and not (str.isalnum(c) or c in ('-', '_', '.')):
                key = current
                current = '' if c == ':' else c
                whitespace = ''
            else:
                # Any other character. Include whitespace too if it was collected
                current += whitespace + c
                whitespace = ''

        if quote:
            # Unterminated quoted string
            raise ValueError("unmatched quotes")

        assert key is None
        assert current == ''
        assert whitespace == ''


def combine_group(left: typing.Optional[float], right: typing.Optional[float]) -> typing.Optional[float]:
    """ Combine group values

    Returns the **weakest** match, ie. the higher numeric value.
    If one value is ``None``, returns the other value.
    """
    if left is None:
        return right
    elif right is None:
        return left
    else:
        return max(left, right)


class Resources:
    """ Manage resources for a task and worker.

    resource value handling:

    boolean items:

    * *tag*: required tag
    * *?tag*: optional tag
    * *~tag*: excluded tag

    The following table indicates the match value for boolean tags of given type,
    :py:class:`~momotor.shared.resources.NoMatch` indicates that the :py:class:`~momotor.shared.resources.NoMatch`
    exception is raised, any other value indicates that that value is returned.

    * `-`: worker or task does not have this tag defined
    * blank cell: ignore, no change in match value

    TODO numeric items

    :param definition: A dictionary of group names to resource groups

    """
    def __init__(self, definition: typing.Dict[str, ResourceGroup] = None):
        self._definition = definition or {}

    # Factories

    @classmethod
    def from_dict(cls: typing.Type[RT], definition: typing.Dict[str, typing.Union[str, typing.Iterable[str]]]) -> RT:
        """ Factory to create a :py:class:`~momotor.shared.resources.Resources` object from a dict
        of group names to group string definitions """
        return cls({
            key: ResourceGroup.create(value) for key, value in definition.items()
        })

    @classmethod
    def from_key_value(cls: typing.Type[RT], key: str, value: typing.Union[str, typing.Iterable[str]]) -> RT:
        """ Factory to create a :py:class:`~momotor.shared.resources.Resources` object from a key and values

        :param key: Group name
        :param value: Group definition string or list of strings
        """
        return cls({
            key: ResourceGroup.create(value)
        })

    @classmethod
    def from_string(cls: typing.Type[RT], value: typing.Union[str, typing.Iterable[str]]) -> RT:
        """ Factory to create a :py:class:`~momotor.shared.resources.Resources` object from
        one or more strings

        :param value: A string or iterable of strings containing resource definitions including a group name
        :return: The parsed resources
        :raises: :py:exc:`ValueError` if the string is an invalid resources definition
        """
        resources = collections.deque()

        if isinstance(value, str):
            resources.extend(split_resources(value))
        else:
            for s in value:
                resources.extend(split_resources(s))

        return cls.union(*(
            cls.from_key_value(key, value) for key, value in resources
        ))

    @classmethod
    def union(cls: typing.Type[RT], *resources: RT) -> RT:
        """ Merge multiple :py:class:`~momotor.shared.resources.Resources` objects into a new one

        :returns: The merged resources
        """

        return cls({
            key: ResourceGroup.union(*(resource.get(key) for resource in resources))
            for key in unique_everseen(
                itertools.chain(*(resource.group_names() for resource in resources))
            )
        })

    # Implementation

    def __len__(self):
        return len(self._definition)

    def group_names(self) -> typing.List[str]:
        """ Returns a list of group names

        :return: group names
        """
        return list(self._definition.keys())

    def get(self, key: str) -> ResourceGroup:
        """ Get the :py:class:`~momotor.shared.resources.ResourceGroup` by name

        :param key: Group name
        :return: The resource group matching the group name
        """
        return self._definition.get(key, ResourceGroup())

    def match(self, worker_resource: "Resources") -> float:
        """ Returns the match between these resources and the provided worker's resources.

        :param worker_resource: The worker's resources
        :return: match value, the lower the value, the better the match is (ranges from
                 -\ :py:data:`~math.inf` to +\ :py:data:`~math.inf`).
                 If nothing matches, but also nothing is excluded, will return
                 :py:data:`~momotor.shared.resources.const.NEUTRAL`
        :raises: :py:exc:`~momotor.shared.resources.NoMatch` if no match can be made (eg. due to an exclude)
        """
        match = None

        group_names = set(self.group_names()) | set(worker_resource.group_names())
        for name in group_names:
            group_match = self.get(name).match(worker_resource.get(name))
            match = combine_group(match, NEUTRAL if group_match is None else group_match)

        return NEUTRAL if match is None else match

    def as_tuples(self) -> typing.Iterable[typing.Tuple[str, typing.Tuple[str]]]:
        for key, group in self._definition.items():
            yield key, group.as_str_tuple()

    def as_str_tuples(self) -> typing.Iterable[typing.Tuple[str, str]]:
        for key, group in self._definition.items():
            yield key, group.as_str()

    def as_str(self, *, multiline: bool = False, compact: bool = False) -> str:
        """ Convert these resources into a string that can be parsed back into a
        :py:class:`~momotor.shared.resources.Resources` object using
        :py:meth:`~momotor.shared.resources.Resources.from_string`

        :param multiline: if True, use newlines to split groups, otherwise uses semicolons
        :param compact: if True, returns a compact representation, otherwise adds more whitespace
        :return: string representation of the resources
        """
        line_sep = '\n' if multiline else (';' if compact else '; ')
        key_sep = ':' if compact else ': '
        value = line_sep.join(f'{quote_str(key)}{key_sep}{group}'.strip() for key, group in self.as_str_tuples())
        if value and multiline and not compact:
            return value + '\n'
        return value

    def __str__(self) -> str:
        return '{' + self.as_str() + '}'



import itertools
import typing

from .exception import NoMatch
from .item import ResourceItem
from .utils import unique_everseen, quote_str, escape_str


UNSAFE_CHARS = {',', ':', ';', '"', "'"}


def quote_item(item: str) -> str:
    if item.startswith(' ') or item.endswith(' ') or '\t' in item or '\n' in item:
        return quote_str(item)

    return escape_str(item, UNSAFE_CHARS)


def combine_item(left: typing.Optional[float], right: typing.Optional[float]) -> typing.Optional[float]:
    """ Combine item values

    Returns the **strongest** match, ie. the lower numeric value.
    If one value is ``None``, returns the other value.
    """
    if left is None:
        return right
    elif right is None:
        return left
    else:
        return min(left, right)


RGT = typing.TypeVar('RGT', bound='ResourceGroup')


class ResourceGroup:
    """ A group of :py:class:`~momotor.shared.resources`.
    """
    def __init__(self, items: typing.Iterable[ResourceItem] = None):
        self._items = tuple(items) if items is not None else tuple()

    # Factories

    @classmethod
    def create(cls: typing.Type[RGT], value: typing.Union[str, typing.Iterable[str]]) -> RGT:
        return cls(ResourceItem.create(value))

    @classmethod
    def union(cls: typing.Type[RGT], *groups: RGT) -> RGT:
        """ Merge multiple resource groups into a new one """
        return cls(unique_everseen(itertools.chain(*(group.items for group in groups))))

    # Implementation

    def __len__(self):
        return len(self._items)

    @property
    def items(self) -> typing.Tuple[ResourceItem]:
        return self._items

    def match(self, worker_group: typing.Optional["ResourceGroup"]) -> typing.Optional[float]:
        """ Returns the match between this resource group and the provided worker's resource group

        :param worker_group: The worker's resource group
        :return: match value, the lower the value, the better the match is (ranges from
                 -\ :py:data:`~math.inf` to +\ :py:data:`~math.inf`)
        :raises: :py:exc:`~momotor.shared.resources.NoMatch` if there are missing or excluded tags
        """
        match = None
        unmatched = frozenset(worker_group.items) if worker_group is not None else frozenset()
        for task_resource in self.items:
            matched = set()
            for worker_resource in unmatched:
                if task_resource.comparable(worker_resource):
                    matched.add(worker_resource)
                    match = combine_item(match, task_resource.compare(worker_resource))

            # If no tag matched, but the task tag is required, it's no match
            if not matched and task_resource.required:
                raise NoMatch

            unmatched -= matched

        # process resources defined on the worker but not the task
        for worker_resource in unmatched:
            match = combine_item(match, worker_resource.compare_missing(worker_resource))

        return match

    def as_str_tuple(self) -> typing.Tuple[str]:
        return tuple(item.as_str() for item in self._items)

    def as_str(self) -> str:
        return ','.join(quote_item(item) for item in self.as_str_tuple())



from asyncio import wrap_future
from concurrent.futures import ThreadPoolExecutor, Future
from logging import getLogger, Logger, DEBUG, INFO, WARNING, ERROR, CRITICAL, raiseExceptions, LogRecord

import sys
import typing
import warnings
from contextlib import asynccontextmanager, contextmanager

__all__ = ['AsyncLogWrapper', 'getAsyncLogger', 'async_log_exception', 'log_exception']

DEFAULT_WAIT_FOR_COMPLETION = False
_executor = ThreadPoolExecutor(max_workers=1)
_last_future: typing.Optional[Future] = None


# noinspection PyPep8Naming
class AsyncLogWrapper:
    """ Wraps a Python logger for async use.
    The original logger is available on the `sync` attribute for use in synchronous code.

    Most properties and methods from the original logger are proxied.

    :param logger: The Python logger to wrap
    :param wait_for_completion: If `True` all logging statements will only return when the line is written to the log.
                                If `False` will queue the line to be logged and return immediately.
                                If `None` (default) will wait if the level of the logger is DEBUG
    """
    # noinspection PyProtectedMember
    def __init__(self, logger: Logger, *, wait_for_completion: typing.Optional[bool] = DEFAULT_WAIT_FOR_COMPLETION):
        self._logger = logger
        self._sync_log = logger._log
        self._wait_for_completion: typing.Optional[bool] = logger.isEnabledFor(DEBUG) \
            if wait_for_completion is None else wait_for_completion

    @property
    def name(self):
        """ Get the name of the logger """
        return self._logger.name

    @property
    def level(self):
        """ Get the level of the logger """
        return self._logger.level

    @property
    def propagate(self):
        """ Get/set the `propagate` property of the logger """
        return self._logger.propagate

    @propagate.setter
    def propagate(self, value):
        self._logger.propagate = value

    @property
    def disabled(self):
        """ Get/set the `disabled` property of the logger """
        return self._logger.disabled

    @disabled.setter
    def disabled(self, value):
        self._logger.disabled = value

    def setLevel(self, level):
        """
        Set the logging level of this logger.  level must be an int or a str.
        """
        return self._logger.setLevel(level)

    def getEffectiveLevel(self):
        """
        Get the effective level for this logger.

        Loop through this logger and its parents in the logger hierarchy,
        looking for a non-zero logging level. Return the first one found.
        """
        return self._logger.getEffectiveLevel()

    def isEnabledFor(self, level):
        """
        Is this logger enabled for level 'level'?
        """
        return self._logger.isEnabledFor(level)

    async def debug(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'DEBUG'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.debug("Houston, we have a %s", "thorny problem", exc_info=1)
        """
        if self.isEnabledFor(DEBUG):
            await self._log(DEBUG, msg, args, **kwargs)

    async def info(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'INFO'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.info("Houston, we have a %s", "interesting problem", exc_info=1)
        """
        if self.isEnabledFor(INFO):
            await self._log(INFO, msg, args, **kwargs)

    async def warning(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'WARNING'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.warning("Houston, we have a %s", "bit of a problem", exc_info=1)
        """
        if self.isEnabledFor(WARNING):
            await self._log(WARNING, msg, args, **kwargs)

    async def warn(self, msg, *args, **kwargs):
        warnings.warn("The 'warn' method is deprecated, use 'warning' instead", DeprecationWarning, 2)
        await self.warning(msg, *args, **kwargs)

    async def error(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'ERROR'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.error("Houston, we have a %s", "major problem", exc_info=1)
        """
        if self.isEnabledFor(ERROR):
            await self._log(ERROR, msg, args, **kwargs)

    async def exception(self, msg, *args, exc_info=True, **kwargs):
        """
        Convenience method for logging an ERROR with exception information.
        """
        await self.error(msg, *args, exc_info=exc_info, **kwargs)

    async def critical(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'CRITICAL'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.critical("Houston, we have a %s", "major disaster", exc_info=1)
        """
        if self.isEnabledFor(CRITICAL):
            await self._log(CRITICAL, msg, args, **kwargs)

    fatal = critical

    async def log(self, level, msg, *args, **kwargs):
        """
        Log 'msg % args' with the integer severity 'level'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.log(level, "We have a %s", "mysterious problem", exc_info=1)
        """
        if not isinstance(level, int):
            if raiseExceptions:
                raise TypeError("level must be an integer")
            else:
                return

        if self.isEnabledFor(level):
            await self._log(level, msg, args, **kwargs)

    @staticmethod
    async def flush():
        """ Wait until the last submitted logging statement has completed
        """
        global _last_future

        fut: typing.Optional[Future] = _last_future
        _last_future = None
        if fut:
            # noinspection PyBroadException
            try:
                await wrap_future(fut)
            except:
                pass

    @staticmethod
    def flush_sync():
        """ Wait until the last submitted logging statement has completed
        """
        global _last_future

        fut: typing.Optional[Future] = _last_future
        _last_future = None
        if fut is not None:
            # noinspection PyBroadException
            try:
                fut.result()
            except:
                pass

    async def _log(self, level, msg, args, exc_info=None, **kwargs):
        global _last_future, _executor

        if args:
            msg = msg % args

        if exc_info:
            if isinstance(exc_info, BaseException):
                exc_info = (type(exc_info), exc_info, exc_info.__traceback__)
            elif not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()

        wait_for_completion = kwargs.pop(
            'wait_for_completion',
            self._wait_for_completion or level >= CRITICAL or bool(exc_info)
        )

        _last_future = _executor.submit(self._sync_log, level, msg, [], exc_info=exc_info, **kwargs)

        # Wait until the message has been processed if `wait_for_completion` is True, or if the
        # log message is critical or has exception info
        if wait_for_completion:
            await self.flush()

    async def handle(self, record: LogRecord):
        """
        Call the handlers for the specified record.

        This method is used for unpickled records received from a socket, as
        well as those created locally. Logger-level filtering is applied.
        """
        global _last_future, _executor

        if (not self.disabled) and self._logger.filter(record):
            wait_for_completion = self._wait_for_completion or record.levelno >= CRITICAL or bool(record.exc_info)

            _last_future = _executor.submit(self._logger.callHandlers, record)

            # Wait until the message has been processed if `wait_for_completion` is True, or if the
            # log message is critical or has exception info
            if wait_for_completion:
                await self.flush()

    @property
    def sync(self) -> Logger:
        """
        Get the original logger instance, for use in synchronous code, e.g.

        logger.sync.info("This is logged from synchronous code")
        """
        # Wait until all async log messages have been processed
        self.flush_sync()

        return self._logger


# noinspection PyPep8Naming
def getAsyncLogger(name, *, wait_for_completion: typing.Optional[bool] = DEFAULT_WAIT_FOR_COMPLETION) \
        -> AsyncLogWrapper:

    """ Convenience function to create an asynchronous logger.
    Creates the logger by calling :py:func:`logging.getLogger` and wraps it with
    :py:class:`~momotor.shared.log.AsyncLogWrapper`

    :param name: name of the logger
    :param wait_for_completion: If `True` all logging statements will only return when the line is written to the log.
                                If `False` will queue the line to be logged and return immediately.
                                If `None` (default) will wait if the level of the logger is DEBUG
    """
    return AsyncLogWrapper(getLogger(name), wait_for_completion=wait_for_completion)


def is_iterable(x):
    try:
        iter(x)
    except TypeError:
        return False
    else:
        return True


@asynccontextmanager
async def async_log_exception(async_logger: AsyncLogWrapper, msg: str, *args,
                              reraise: bool = False,
                              ignore: typing.Union[Exception, typing.Sequence[Exception], None] = None):
    """ An async context manager that captures exceptions, logs them, and optionally re-raises them.

    :param async_logger: The async logger to log the exception to
    :param msg: A message to add to the exception
    :param args: Any optional arguments for the message
    :param reraise: If `True`, the exception will be re-raised after logging
    :param ignore: An exception, or list of exceptions, to ignore. These exceptions will not be logged but always
                   re-raised
    """
    # noinspection PyBroadException
    try:
        yield
    except Exception as e:
        if ignore and isinstance(e, tuple(ignore) if is_iterable(ignore) else ignore):
            raise

        await async_logger.exception(msg, *args)
        if reraise:
            raise


@contextmanager
def log_exception(logger: Logger, msg: str, *args, reraise=False, ignore=None):
    """ A context manager that captures exceptions, logs them, and optionally re-raises them.

    :param logger: The logger to log the exception to
    :param msg: A message to add to the exception
    :param args: Any optional arguments for the message
    :param reraise: If `True`, the exception will be re-raised after logging
    :param ignore: An exception, or list of exceptions, to ignore. These exceptions will not be logged but always
                   re-raised
    """
    # noinspection PyBroadException
    try:
        yield
    except Exception as e:
        if ignore and isinstance(e, tuple(ignore) if is_iterable(ignore) else ignore):
            raise

        logger.exception(msg, *args)
        if reraise:
            raise



import typing

from .const import STRONGEST, STRONG, WEAK, WEAKEST
from .exception import NoMatch

PRIO_MAPPING = {
    (True, True): STRONGEST,
    (True, False): STRONG,
    (False, True): WEAK,
    (False, False): WEAKEST,
}


def split_items(items: str) -> typing.Generator[str, None, None]:
    current = ''  # Current value being collected
    whitespace = ''  # Whitespace that might or might not be part of the value
    quote = None  # None: not inside a quoted string, otherwise the quote itself
    escape = False  # True: include next character unconditionally

    for c in items+',':
        if escape:
            # This character is escaped, so include it unconditionally
            escape = False
            current += c
        elif c == '\\':
            # Escape next character
            escape = True
        elif quote:
            # Inside a quoted string
            if c == quote:
                # End of string
                quote = None
            else:
                # Include any other character
                current += c
        elif c in (' ', '\t'):
            # Whitespace. Ignore leading whitespace, collect all other whitespace in a separate variable
            if current:
                whitespace += c
        elif c in ('"', "'"):
            # Start of a quoted string. Also include whitespace into current
            quote = c
            current += whitespace
            whitespace = ''
        elif c == ',':
            # Field separator. Ignore trailing whitespace, reset current value
            yield current
            current = ''
            whitespace = ''
        elif c == ':':
            # An unescaped colon is not allowed (See #3)
            raise ValueError("colons in items must be escaped")
        else:
            # Any other character. Include whitespace too if it was collected
            current += whitespace + c
            whitespace = ''

    if quote:
        # Unterminated quoted string
        raise ValueError("unmatched quotes")

    assert current == ''
    assert whitespace == ''


RIT = typing.TypeVar('RIT', bound='ResourceItem')


class ResourceItem:
    def __init__(self, value: str, required: bool, excluded: bool):
        if required and excluded:
            raise ValueError('cannot combine required and excluded flags')

        self.value = value
        self.required = required
        self.excluded = excluded

    @classmethod
    def _create_item(cls: typing.Type[RIT], value) -> RIT:
        from .tag import Tag
        return Tag(value)

    @classmethod
    def create(cls: typing.Type[RIT], value: typing.Union[str, typing.Iterable[str]]) -> typing.Iterable[RIT]:
        if isinstance(value, str):
            for item in split_items(value):
                if item:
                    yield cls._create_item(item)
        else:
            for element in value:
                for item in split_items(element):
                    if item:
                        yield cls._create_item(item)

    def comparable(self, worker: RIT) -> bool:
        return type(self) == type(worker)

    def compare(self, worker: RIT) -> typing.Optional[float]:
        """
            +-----------+------------------------------------------------------+----------------------------------------------------+-----------------------------------------------+
            | . worker  | required                                             | optional                                           | excluded                                      |
            | task      |                                                      |                                                    |                                               |
            +-----------+------------------------------------------------------+----------------------------------------------------+-----------------------------------------------+
            | required  | :py:data:`~momotor.shared.resources.const.STRONGEST` | :py:data:`~momotor.shared.resources.const.STRONG`  | :py:class:`~momotor.shared.resources.NoMatch` |
            +-----------+------------------------------------------------------+----------------------------------------------------+-----------------------------------------------+
            | optional  | :py:data:`~momotor.shared.resources.const.WEAK`      | :py:data:`~momotor.shared.resources.const.WEAKEST` | :py:class:`~momotor.shared.resources.NoMatch` |
            +-----------+------------------------------------------------------+----------------------------------------------------+-----------------------------------------------+
            | excluded  | :py:class:`~momotor.shared.resources.NoMatch`        | :py:class:`~momotor.shared.resources.NoMatch`      |                                               |
            +-----------+------------------------------------------------------+----------------------------------------------------+-----------------------------------------------+
        """
        if self.comparable(worker):
            if self.excluded or worker.excluded:
                if self.excluded and worker.excluded:
                    return None
                else:
                    raise NoMatch

            return PRIO_MAPPING[(self.required, worker.required)]

        return None

    @classmethod
    def compare_missing(cls: typing.Type[RIT], worker: RIT) -> typing.Optional[float]:
        """
            +----------+-----------------------------------------------+----------------------------------------------------+------------------------------------------------+
            | . worker | required                                      | optional                                           | excluded                                       |
            | task     |                                               |                                                    |                                                |
            +----------+-----------------------------------------------+----------------------------------------------------+------------------------------------------------+
            | `-`      | :py:class:`~momotor.shared.resources.NoMatch` |                                                    |                                                |
            +----------+-----------------------------------------------+----------------------------------------------------+------------------------------------------------+
        """
        if worker.required:
            raise NoMatch

        return None

    def as_str(self) -> str:
        if self.excluded:
            prefix = '~'
        elif self.required:
            prefix = ''
        else:
            prefix = '?'

        return prefix + self.value



from abc import ABC
from contextlib import asynccontextmanager
import typing


class StateABC(ABC):
    """ Abstract base class to implement shared locks
    """
    @asynccontextmanager
    async def get_lock(self, key: str, *, exclusive: bool = True) -> typing.AsyncContextManager[None]:
        """ Context manager to get a lock. The lock is held while the context is active.

        The lock is an 'exclusive' or 'read/write' lock:

        * Many processes can hold the lock with ``exclusive``\ ==\ ``False`` (*read* mode), but
        * Only one process can hold the lock with ``exclusive``\ ==\ ``True`` (*write* mode).
          No other processes can hold the lock while one process has it locked exclusively.

        :param key: global name for the lock
        :param exclusive: get an exclusive lock
        :return:
        :raises: :py:class:`~momotor.shared.state.LockFailed`
        """
        raise NotImplementedError

    async def test_lock(self, key: str, *, exclusive: bool = True) -> bool:
        """ Test the lock. Returns ``True`` if the lock is currently not being held.

        Not that this is very transient information. The lock could be acquired immediately after this method
        returns.

        :param key: global name for the lock
        :param exclusive: get an exclusive lock
        :return:
        """
        raise NotImplementedError



from itertools import filterfalse
import re
import typing


def unique_everseen(iterable, key=None):
    # https://docs.python.org/3.7/library/itertools.html#itertools-recipes
    """List unique elements, preserving order. Remember all elements ever seen."""
    # unique_everseen('AAAABBBCCDAABBB') --> A B C D
    # unique_everseen('ABBCcAD', str.lower) --> A B C D
    seen = set()
    seen_add = seen.add
    if key is None:
        for element in filterfalse(seen.__contains__, iterable):
            seen_add(element)
            yield element
    else:
        for element in iterable:
            k = key(element)
            if k not in seen:
                seen_add(k)
                yield element


SAFE_STR_RE = re.compile(r'\w+')


def escape_str(s: str, escape_chars: typing.Set[str] = None) -> str:
    """ Return string with characters in `escape_chars` escaped.
    Uses the backslash character for escapes, so that is always escaped and does not have to be supplied
    to `escape_chars`
    """
    escape_chars = {'\\'} | (escape_chars if escape_chars else set())

    result = ''
    for c in s:
        if c in escape_chars:
            result += '\\'

        result += c

    return result


def quote_str(s: str, ) -> str:
    """ Quote string containing unsafe characters for both group-name and items
    :param s: string to make safe
    :return:
    """
    # Only containing safe characters
    m = SAFE_STR_RE.match(s)
    if m and m[0] == s:
        return escape_str(s)

    has_single_quotes = "'" in s
    has_double_quotes = '"' in s

    if has_single_quotes and has_double_quotes:
        return '"' + escape_str(s, {'"'}) + '"'
    elif has_single_quotes:
        return '"' + escape_str(s) + '"'
    else:
        return "'" + escape_str(s) + "'"



import math

#: Match value for tags. ``STRONGEST`` indicates the best possible match: both task and worker require this tag
STRONGEST = -math.inf

#: Match value for tags. ``STRONG`` indicates a good match: the task requires this tag, for the worker it is optional
STRONG = -1.0

#: Match value for tags. ``NEUTRAL`` indicates there are no matching tags between task and worker.
NEUTRAL = 0.0

#: Match value for tags. ``WEAK`` indicates a slightly worst match: the worker requires the tag,
#: for the task it is optional
WEAK = 1.0

#: Match value for tags. ``WEAKEST`` indicates a worst possible match: for both worker and task the tag is optional
WEAKEST = math.inf



import aiorwlock

import typing


class ExLock(aiorwlock.RWLock):
    """
    A subclass of `aiorwlock.RWLock <https://github.com/aio-libs/aiorwlock>`_

    An :py:class:`~momotor.shared.exlock.ExLock` maintains a pair of associated locks, one for read-only
    operations and one for writing. The read lock may be held simultaneously
    by multiple reader tasks, so long as there are no writers. The write
    lock is exclusive.
    """

    # noinspection PyProtectedMember
    def get(self, exclusive: bool) -> typing.Union[aiorwlock._ReaderLock, aiorwlock._WriterLock]:
        """ Get either the :py:attr:`reader` or :py:attr:`writer` lock, depending on the value of
        ``exclusive``

        :param exclusive: if ``False``, requests the :py:attr:`reader` lock, otherwise the :py:attr:`writer` lock
        :return: the requested lock
        """
        return self.writer if exclusive else self.reader



from .item import ResourceItem


class Tag(ResourceItem):
    def __init__(self, value: str):
        if value.startswith('?'):
            value = value[1:]
            required, excluded = False, False
        elif value.startswith('~'):
            value = value[1:]
            required, excluded = False, True
        else:
            required, excluded = True, False

        super().__init__(value, required, excluded)

    def comparable(self, worker: ResourceItem) -> bool:
        return super().comparable(worker) and self.value == worker.value



__all__ = ['annotate_docstring']


def annotate_docstring(*args, **kwargs):
    """ Decorator to help write better docstrings.

    The arguments of the decorator are applied using :py:meth:`str.format`
    to the __doc__ of the class or method decorated.

    Example:

    .. code:: python

       logger = getLogger('some.logger')

       @annotate_docstring(logger=logger)
       class SomeClass:
          \"\"\" This class does some logging on the "{logger.name}" logger
          \"\"\"

    In the example above, the docstring will read:

    .. code:: text

       This class does some logging on the "some.logger" logger

    :param args: the arguments for :py:meth:`str.format`
    :param kwargs: the keyword arguments for :py:meth:`str.format`
    :return: the decorated object with annotated doc string
    """

    def decorator(klass):
        if klass.__doc__ is not None:
            klass.__doc__ = klass.__doc__.format(*args, **kwargs)
        return klass

    return decorator



import os
from setuptools import setup, find_namespace_packages


def get_version():
    import os.path

    here = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(here, 'src', 'momotor', 'shared', 'version.py'), 'r') as version_file:
        loc = {}
        exec(version_file.readline(), {}, loc)
        return loc['__VERSION__']


def get_long_description():
    with open("README.md", "r") as fh:
        return fh.read()


full_version = get_version() + os.environ.get('VERSION_TAG', '')
is_dev = '.dev' in full_version

setup(
    name='momotor-engine-shared',
    version=get_version(),
    author='Erik Scheffers',
    author_email='e.t.j.scheffers@tue.nl',
    description="Momotor Engine shared code",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url='https://momotor.org/',
    project_urls={
        'Documentation': f'https://momotor.org/doc/engine/momotor-engine-shared/{"dev" if is_dev else "release"}/{full_version}/',
        'Source': 'https://gitlab.tue.nl/momotor/engine-py3/momotor-engine-shared/',
        'Tracker': 'https://gitlab.tue.nl/momotor/engine-py3/momotor-engine-shared/issues',
    },
    install_requires=[
        'aiorwlock',
    ],
    extras_require={
        'test': [
            'pytest',
            'pytest-cov',
            'pytest-pythonpath',
        ],
        'docs': [
            'Sphinx',
            'sphinx-autodoc-typehints',
        ]
    },
    python_requires='>=3.7',
    packages=find_namespace_packages(where='src', include=('momotor.*',)),
    package_dir={'': 'src'},
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    ],
)



import asyncio
from contextlib import asynccontextmanager
import typing
import weakref

from momotor.shared.exlock import ExLock
from .base import StateABC
from ..doc import annotate_docstring
from ..log import getAsyncLogger

logger = getAsyncLogger(__name__)


@annotate_docstring(logger=logger)
class LocalState(StateABC):
    """ Reference implementation of :py:class:`~momotor.shared.state.StateABC` for local use.

    Uses a :py:class:`~weakref.WeakValueDictionary` of :py:class:`momotor.shared.exlock.ExLock` objects.

    Produces debug logging information on the ``{logger.name}`` logger
    """
    def __init__(self):
        self.__dict_lock = asyncio.Lock()
        self.__locks: typing.MutableMapping[str, ExLock] = weakref.WeakValueDictionary()

    async def _get_rwlock(self, key: str, create: bool):
        async with self.__dict_lock:
            lock = self.__locks.get(key)
            if lock is None and create:
                lock = self.__locks[key] = ExLock()

        return lock

    @asynccontextmanager
    async def get_lock(self, key: str, *, exclusive: bool = True) -> typing.AsyncContextManager[None]:
        # Need to keep a reference to the full rwlock, not just the reader or writer lock, otherwise
        # the WeakValueDictionary will delete the rwlock even though we have the reader or writer locked
        # using async-with causes similar problems, therefor using acquire and release

        exlock = await self._get_rwlock(key, True)
        await logger.debug(f'state lock {key}{" exclusive" if exclusive else ""} acquiring {exlock}')

        await exlock.get(exclusive).acquire()
        await logger.debug(f'state lock {key}{" exclusive" if exclusive else ""} acquired {exlock}')

        try:
            yield
        finally:
            exlock.get(exclusive).release()
            await logger.debug(f'state lock {key}{" exclusive" if exclusive else ""} released {exlock}')

    async def test_lock(self, key: str, *, exclusive: bool = True) -> bool:
        exlock = await self._get_rwlock(key, False)
        return exlock.get(exclusive).locked if exlock else False
